#ifndef _PROTO_GRAPH
#define _PROTO_GRAPH

#include <protoTree.h>
#include <protoDebug.h>

class ProtoGraph
{
    public:
            
        class Vertice;
    
        ProtoGraph();
        ~ProtoGraph();

        void InsertVertice(Vertice& vertice);
        void RemoveVertice(Vertice& vertice);

        class Edge
        {
            friend class Vertice;
            
            public:
                // Note: Constructing an Edge automatically bidirectionally connects
                //       the pair of Vertices
                // (TBD) extend to optionally allow for "unidirectional" as well
                Edge(Vertice& a, Vertice& b);
                virtual ~Edge();

                Vertice& GetDst(const Vertice& src) const
                    {return ((&src == &term_a) ? term_b : term_a);}
                
            protected:    
                Edge* GetNext(const Vertice& vertice) const
                {
                    ASSERT((&vertice == &term_a) || (&vertice == &term_b));
                    return ((&vertice == &term_a) ? next_a : next_b);
                }
                
                void AppendToVertices();
                void DetachFromVertices();
                
                void Prepend(const Vertice& vertice, Edge* edge)
                {
                    ASSERT((&vertice == &term_a) || (&vertice == &term_b));
                    if (&vertice == &term_a)
                        prev_a = edge;
                    else
                        prev_b = edge;
                }
                void Append(const Vertice& vertice, Edge* edge)
                {
                    ASSERT((&vertice == &term_a) || (&vertice == &term_b));
                    if (&vertice == &term_a)
                        next_a = edge;
                    else
                        next_b = edge;
                }

                Vertice&    term_a;
                Edge*       prev_a;
                Edge*       next_a;

                Vertice&    term_b;
                Edge*       prev_b;
                Edge*       next_b;

        };  // end class ProtoGraph::Edge

        class Vertice
        {
            friend class ProtoGraph;
            friend class Edge;
            
            public:
                virtual ~Vertice();    

                class EdgeIterator;
                friend class EdgeIterator;
                class EdgeIterator
                {
                    public:
                        EdgeIterator(const Vertice& theVertice);
                        ~EdgeIterator();

                        void Reset()
                            {next_edge = vertice.GetHeadEdge();}

                        Edge* GetNextEdge()
                        {
                            Edge* edge = next_edge;
                            next_edge = vertice.GetNextEdge(edge);
                            return edge;   
                        }

                    private:
                         const Vertice& vertice;
                         Edge*          next_edge;   

                };  // end class Vertice::EdgeIterator

                // Base class for Queues that wish to contain Vertices
                class Queue
                {
                    public:
                        class State
                        {
                            public:
                                virtual ~State();
                                Queue* GetQueue() const
                                    {return queue;}
                                void SetQueue(Queue* theQueue)
                                    {queue = theQueue;}
                                
                                Vertice* GetVertice() const
                                    {return vertice;}
                                void SetVertice(Vertice* theVertice)
                                    {vertice = theVertice;}
                                
                                // container used by Vertices to keep their lists 
                                // of Queue::State
                                class Entry : public ProtoTree::Item
                                {
                                    public:
                                        Entry(State& theState);
                                        State& GetQueueState() const
                                            {return queue_state;}
                                        
                                    private:
                                        State&    queue_state;
                                }; // end class ProtoGraph::Vertice::Queue::State::Entry
                                const Entry& GetEntry() const
                                    {return entry;}
                                
                                Entry& AccessEntry() 
                                    {return entry;}
                                
                            protected:
                                State();
                            
                            
                            private:
                                Entry       entry;
                                Queue*      queue;
                                Vertice*    vertice;
                        };  // end class ProtoGraph::Vertice::Queue::State
                        
                    protected:
                        State* GetQueueState(const Vertice& vertice) const
                            {return vertice.GetQueueState(*this);}
                        void Associate(State& queueState, Vertice& vertice)
                            {vertice.Reference(queueState);}
                        void Disassociate(State& queueState, Vertice& vertice)
                            {vertice.Dereference(queueState);} 
                };  // end class Queue()
                friend class Queue;
                
                bool IsInQueue(const Queue& queue) const
                    {return (NULL != GetQueueState(queue));}
                
                // Simple queue class used for traversal purposes
                class SimpleQueue : public Queue
                {
                    public:
                        class ItemPool;
                    
                        SimpleQueue(SimpleQueue::ItemPool* itemPool = NULL);
                        ~SimpleQueue();
                        
                        bool Prepend(Vertice& vertice);
                        bool Append(Vertice& vertice);
                        void Remove(Vertice& vertice);
                        
                        bool IsEmpty() const
                            {return (NULL == head);}
                        
                        void Empty() 
                            {while (!IsEmpty()) RemoveHead();}
                        
                        Vertice* GetHead() const
                            {return ((NULL != head) ? head->GetVertice() : NULL);}
                        
                        Vertice* RemoveHead()
                        {
                            if (NULL != head)
                            {
                                Item* item = head;
                                Vertice* vertice = item->GetVertice();
                                ASSERT(NULL != vertice);
                                RemoveItem(*item);
                                Disassociate(*item, *vertice);
                                if (NULL != item_pool)
                                    item_pool->Put(item);
                                else
                                    delete item;
                                return vertice; 
                            }
                            else
                            {
                                return NULL;
                            }  
                        }
                        
                        Vertice* GetNext(const Vertice& vertice) const
                        {
                            Item* item = static_cast<Item*>(vertice.GetQueueState(*this));
                            return ((NULL != item) ? item->GetVertice() : NULL);
                        }
                        
                        // Move vertice from this queue to another
                        void Transfer(Vertice& vertice, SimpleQueue& dstQueue)
                        {
                            Item* item = static_cast<Item*>(vertice.GetQueueState(*this));
                            ASSERT(NULL != item);
                            RemoveItem(*item); // item still references this, vertice
                            // Put this stuff in RemoveItem()/AppendItem() methods
                            Disassociate(*item, vertice);
                            item->SetQueue(&dstQueue);
                            Associate(*item, vertice);
                            // AppendItem() changes item to reference dstQueue
                            dstQueue.AppendItem(*item); 
                              
                        }
                        
                    protected:
                        class Item;
                    
                    public:
                   
                        friend class ItemPool;
                        class ItemPool
                        {
                            friend class SimpleQueue;
                            public:
                                ItemPool();
                                ~ItemPool();
                                bool IsEmpty() const
                                    {return (NULL == head);}
                                void Destroy();
                                
                            private:
                                Item* Get();
                                void Put(Item* item)
                                {
                                    item->Append(head);
                                    head = item;
                                }    
                                
                                Item* head;
                        };  // end class ProtoGraph::Vertice::SimpleQueue::ItemPool
                        
                        
                    protected:
                        class Item : public Queue::State
                        {
                            public:
                                Item();
                                ~Item();

                                void Prepend(Item* theItem) 
                                    {prev = theItem;}
                                void Append(Item* theItem) 
                                    {next = theItem;}
                                
                                Item* GetPrev() const 
                                    {return prev;}
                                Item* GetNext() const
                                    {return next;}

                            private:
                                Item*       prev;    
                                Item*       next;
                        };  // end class ProtoGraph::Vertice::SimpleQueue::Item
                    
                    public:    
                        class Iterator;
                        friend class Iterator;
                        class Iterator
                        {
                            public:
                                Iterator(const SimpleQueue& theQueue, bool reverse = false);
                                ~Iterator();

                                void Reset();
                                Vertice* GetNextVertice();
                                
                            private:
                                const SimpleQueue&  queue;
                                Item*               next_item;
                                bool                forward;

                        };  // end class Vertice::SimpleQueue::Iterator

                    protected:
                        Item* GetNewItem()
                            {return ((NULL != item_pool) ? item_pool->Get() : new Item);}
                        void AppendItem(Item& item);
                        void PrependItem(Item& item);
                        void RemoveItem(Item& item);
                            
                        Item*       head;
                        Item*       tail;
                        ItemPool*   item_pool;
                        
                };  // end class Vertice::SimpleQueue()
                
                
            protected:
                Vertice();
            
            private:
                Edge* GetNextEdge(Edge* edge) const
                    {return (edge ? edge->GetNext(*this) : NULL);}
                
                Edge* GetHeadEdge() const {return edge_list_head;}
                void SetHeadEdge(Edge* edge) {edge_list_head = edge;}
                Edge* GetTailEdge() const {return edge_list_tail;}
                void SetTailEdge(Edge* edge) {edge_list_tail = edge;}

                void GraphPrepend(Vertice* vertice) {graph_prev = vertice;}
                void GraphAppend(Vertice* vertice) {graph_next = vertice;}
                Vertice* GetGraphPrev() {return graph_prev;}
                Vertice* GetGraphNext() {return graph_next;}
                
                Queue::State* GetQueueState(const Queue& queue) const;
                void Reference(Queue::State& queueState);
                void Dereference(Queue::State& queueState);

                // List of edges (TBD - could this be singly-linked instead?)
                Edge*           edge_list_head;
                Edge*           edge_list_tail;
                
                // These members are for use by traversals and 
                // other queue manipulations as needed
                //Queue::State*   queue_state_head; 
                ProtoTree       queue_state_tree;
                
                // ProtoGraph keeps a linked list of all vertices using these
                Vertice*        graph_prev;
                Vertice*        graph_next;

        };  // end class ProtoGraph::Vertice

        class VerticeIterator ;
        friend class VerticeIterator;
        class VerticeIterator 
        {
            public:
                VerticeIterator(const ProtoGraph& theGraph);
                ~VerticeIterator();

                void Reset();
                Vertice* GetNextVertice()
                {
                    Vertice* vertice = next_vertice;
                    next_vertice = graph.GetNextVertice(vertice);
                    return vertice;   
                }
                
            private:
                const ProtoGraph&   graph;
                Vertice*            next_vertice;
                
        };  // end class ProtoGraph::VerticeIterator
        
        
        // (TBD) We may need to make a separate "SimpleTraversal" base class
        // that SimpleSimpleTraversal (and others) derive from so that a
        // ProtoGraph can inform associated SimpleTraversals if the 
        // graph state changes ???

        class SimpleTraversal
        {
            public:
                SimpleTraversal(const ProtoGraph& theGraph, 
                          Vertice&          startVertice,
                          bool              depthFirst = false);
                ~SimpleTraversal();
                bool Reset();
                Vertice* GetNextVertice(unsigned int* level = NULL);

            protected:
                // Members
                const ProtoGraph&               graph;
                Vertice&                        start_vertice;
                bool                            depth_first;    // false == breadth-first search
                unsigned int                    current_level;
                Vertice*                        trans_vertice;  // level transition marker
                Vertice::SimpleQueue            queue_pending;
                Vertice::SimpleQueue            queue_visited;
                
                Vertice::SimpleQueue::ItemPool  item_pool;
        };  // end class ProtoGraph::SimpleTraversal

    private:
        Vertice* GetPrevVertice(Vertice* vertice) const
            {return (vertice ? vertice->GetGraphPrev() : NULL);} 
        Vertice* GetNextVertice(Vertice* vertice) const
            {return (vertice ? vertice->GetGraphNext() : NULL);}    
            
        Vertice*    head_vertice;  
        Vertice*    tail_vertice; 
            
};  // end class ProtoGraph



#endif // _PROTO_GRAPH
