#ifndef _MANET_GRAPH
#define _MANET_GRAPH

#include "protoGraph.h"
#include <protoAddress.h>
#include <arpa/inet.h>  // for ntohl(), etc
#include <sys/types.h>  // for BYTE_ORDER macro

// IMPORTANT NOTE: UNDER CONSTRUCTION!!!!  (WORK IN PROGRESS)
//
// I'm changing this concurrently.  Likely will be reworking ProtoGraph
// and ManetGraph significantly!!!!


class ManetGraph : public ProtoGraph
{
    public:
        ManetGraph();
        ~ManetGraph();
        
        
        // Base class for the variants of "Cost" we use for
        // various graph traversals, etc
        class Cost
        {
            public:
                virtual ~Cost();
                
                // Required overrides
                virtual unsigned int GetKeyLength() const = 0;
                virtual const char* GetKeyValue() const = 0;
                virtual bool UseSignBit() const = 0;
                virtual bool UseComplement2() const = 0;
                virtual void Minimize()  = 0;
                virtual void Maximize() = 0;
                virtual bool IsMax() const = 0;
                virtual bool IsMin() const = 0;
                virtual Cost* Clone() const = 0;
                virtual bool Copy(const Cost& cost) = 0;
                
                // Cost comparison methods (should we allow override?)
                // (These defaults allow comparison of disparate Cost types)
                bool operator==(const Cost& cost) const
                {
                    bool result = (GetKeyLength() == cost.GetKeyLength());
                    bool cmpr = (0 == memcmp(cost.GetKeyValue(), GetKeyValue(), GetKeyLength()));
                    TRACE("count keylen cmp result:%d cmpr:%d\n", result, cmpr);
                    
                    return (result ? cmpr : false);
                }
                
                bool operator!=(const Cost& cost) const
                {
                    bool result = (cost == *this);
                    return (result ? false : true);
                }
                
                bool operator>(const Cost& cost) const;
                
                bool operator>=(const Cost& cost) const
                    {return ((*this > cost) || (*this == cost));}
            
                bool operator<(const Cost& cost) const
                    { return ((*this >= cost) ? false : true);}
                
                bool operator<=(const Cost& cost) const
                    {return ((*this > cost) ? false : true);}
                
                
            protected:
                Cost();
                
        };  // end class ManetGraph::Cost
                   
        // This enforces a positive (non-negative) "double" floating point cost
        class SimpleCost : public Cost
        {     
            public:
                SimpleCost(bool minimize = true);
                SimpleCost(const SimpleCost& cost);
                ~SimpleCost();
            
                // Required overrides of pure virtual methods
                unsigned int GetKeyLength() const
                    {return (sizeof(double));}
                const char* GetKeyValue() const
                    {return ((char*)&value);}
                bool UseSignBit() const
                    {return true;}
                bool UseComplement2() const
                    {return true;}
                void Minimize() 
                    {value = 0.0;}
                void Maximize() 
                    {value = MAX_VALUE;}
                bool IsMax() const
                    {return MAX_VALUE == value;}
                bool IsMin() const
                    {return 0.0 == value;}
                
                // (TBD) change when "Covariant Return Types" are ubiquitous
                Cost* Clone() const
                    {return (static_cast<Cost*>(new SimpleCost(*this)));}
                bool Copy(const Cost& cost);
                
                // Value manipulation methods
                double GetValue() const {return ntohd(value);}
                void SetValue(double theValue) 
                    {value = (theValue >= 0.0) ? htond(theValue) : MAX_VALUE;}
                
                void operator+=(const SimpleCost& cost);
                
                // Misc
                static double ComputeMaxValue();
                

            private:
                // We use these assuming IEEE754 floating point
                static double ntohd(double val)
                {
# if BYTE_ORDER == LITTLE_ENDIAN
                    double result;
                    long* outPtr = (long*)&result;
                    long* inPtr = (long*)&val;
                    outPtr[0] = ntohl(inPtr[1]);
                    outPtr[1] = ntohl(inPtr[0]);
                    return result;
#else
                    return val;
#endif  // if/else __BIG_ENDIAN
                }
                static double htond(double val)
                {
                    return ntohd(val);
                }
                static const double MAX_VALUE;
            
                double value;  // stored in Big Endian order

        };  // end class ManetGraph::SimpleCost
        
        static const SimpleCost COST_MAX;

        class Interface;

        class Link : public Edge
        {
            public:
                Link(Interface& a, Interface& b);
                ~Link();
                
                Interface& GetDst(const Interface& src) const
                    {return static_cast<Interface&>(Edge::GetDst(src));}

                const Cost* GetCost(const Interface& src) const
                {
                    ASSERT((&src == &term_a) || (&src == &term_b));
                    return ((static_cast<const Vertice*>(&src) == &term_a) ? cost_a : cost_b);   
                }
                bool SetCost(const Interface& src, const Cost& cost); 
                bool SetCost(const Cost& cost);

            private:
                bool SetCostInternal(Cost** costHandle, const Cost& cost);
            
                Cost*   cost_a;  // cost to 'term_a' from 'term_b'
                Cost*   cost_b;  // cost to 'term_b' from 'term_a'

        };  // end class ManetGraph::Link

        class Node;
        class Interface : public ProtoGraph::Vertice, public ProtoTree::Item
        {
            
            public:
                Interface(Node& theNode, const ProtoAddress& addr);
                ~Interface();    
                
                const ProtoAddress& GetAddress() const 
                    {return address;}
                void SetAddress(const ProtoAddress& theAddress)
                    {address = theAddress;}

                Node& GetNode() const {return node;}
                
                void Append(Interface* iface) {node_next = iface;}
                Interface* GetNodeNext() const {return node_next;}
                

                // Add/Mod link from this -> iface
                bool Connect(Interface& iface, const Cost& theCost, bool duplex = true);
                void Disconnect(Interface& iface, bool duplex = true);

                class LinkIterator : public Vertice::EdgeIterator
                {
                    public:
                        LinkIterator(Interface& theInterface);
                        ~LinkIterator();

                        void Reset() 
                            {EdgeIterator::Reset();}

                        Link* GetNextLink()
                            {return static_cast<Link*>(GetNextEdge());}
                };  // end class Interface::LinkIterator
                
                
                // Queue class to use for _temporary_ traversal purposes
                class PriorityQueue : public Vertice::Queue
                {
                    public:
                        class ItemPool;
                        PriorityQueue(ItemPool* itemPool = NULL);
                        ~PriorityQueue();
                        
                        //bool IsReady() const
                        //    {return (iface_list.IsReady());}
                        
                        // Ordered by "Cost"
                        bool Insert(Interface& iface, const Cost& cost);
                        void Remove(Interface& iface);
                        void Adjust(Interface& iface, const Cost& newCost);
                        bool AdjustDownward(Interface& iface, const Cost& newCost);
                        
                        void Transfer(Interface& iface, PriorityQueue& dstQueue)
                        {
                            Item* item = static_cast<Item*>(GetQueueState(iface));
                            ASSERT(NULL != item);
                            RemoveItem(*item);
                            Disassociate(*item, iface);
                            item->SetQueue(&dstQueue);
                            Associate(*item, iface);
                            dstQueue.InsertItem(*item);   
                        }
                        
                        bool Prepend(Interface& iface);
                        bool Append(Interface& iface);
                        
                        void Empty();
                        
                        bool IsEmpty() const
                            {return iface_list.IsEmpty();}
                        
                        Interface* GetHead() const
                        {
                            Item* headItem = static_cast<Item*>(iface_list.GetHead()); 
                            return ((NULL != headItem) ? headItem->GetInterface() : NULL);   
                        }
                        
                        const Cost* GetCost(const Interface& iface)
                        {
                            Item* item = static_cast<Item*>(GetQueueState(iface));
                            ASSERT(NULL != item);
                            return (item->GetCost());
                        }
                        
                        
                        void SetRouteLinks(Interface& iface, Link* nextHop, Link* prevHop)
                        {
                            Item* item = static_cast<Item*>(GetQueueState(iface));
                            ASSERT(NULL != item);
                            item->SetNextHopLink(nextHop);
                            item->SetParentLink(prevHop);
                        }
                        
                        Link* GetNextHopLink(const Interface& iface) const
                        {
                            Item* item = static_cast<Item*>(GetQueueState(iface));
                            ASSERT(NULL != item);
                            return (item->GetNextHopLink());
                        }
                        
                        Interface* GetNextHop(const Interface& dstIface, const Interface& srcIface) const;
                        Interface* GetPrevHop(const Interface& dstIface) const;
                        
                        Interface* RemoveHead();
                        
                   
                    protected:
                        class Item;
                    
                    public:     
                        class Iterator : public ProtoSortedTree::Iterator
                        {
                            public:
                                Iterator(PriorityQueue& theQueue, bool reverse = false);
                                ~Iterator();

                                void Reset() 
                                    {ProtoSortedTree::Iterator::Reset();}
                                Interface* GetNextInterface()
                                {
                                    Item* item = static_cast<Item*>(ProtoSortedTree::Iterator::GetNextItem());
                                    return ((NULL != item) ? item->GetInterface() : NULL);
                                }

                        };  // end class ManetGraph::Interface::PriorityQueue::Iterator
                        
                        friend class ItemPool;
                        class ItemPool
                        {
                            friend class PriorityQueue;
                            public:
                                ItemPool();
                                ~ItemPool();
                                bool IsEmpty() const
                                    {return (NULL == head);}
                                void Destroy();
                                
                            private:
                                Item* Get();
                                void Put(Item& item)
                                {
                                    item.SetPoolNext(head);
                                    head = &item;
                                }    
                                Item* head;
                        };  // end class ManetGraph::Interface::PriorityQueue::ItemPool
                        
                        
                    protected:
                        class Item : public Queue::State, public ProtoSortedTree::Item
                        {
                            public:
                                Item();
                                ~Item();
                                
                                void SetInterface(Interface* interface)
                                    {SetVertice(interface);}
                                Interface* GetInterface() const
                                    {return static_cast<Interface*>(GetVertice());}
                                
                                bool SetCost(const Cost& theCost);
                                //void MaximizeCost() {cost.Maximize();}
                                //void MinimizeCost() {cost.Minimize();}
                                const Cost* GetCost() const {return cost;}
                                
                                // These are useful post-Djikstra
                                void SetParentLink(Link* parentLink)
                                    {parent_link = parentLink;}
                                Link* GetParentLink() const {return parent_link;}

                                void SetNextHopLink(Link* nextHopLink) {next_hop_link = nextHopLink;}
                                Link* GetNextHopLink() {return next_hop_link;}

                                void SetLevel(unsigned int theLevel) {level = theLevel;}
                                unsigned int GetLevel() {return level;}
                                       
                            private:
                                Cost*           cost;             // cost to get here from srcIface
                                Link*           parent_link;      // reverse path towards srcIFace
                                Link*           next_hop_link;    // forward path from srcIface to here   
                                unsigned int    level;            // routing tree level
                                
                        };  // end class ManetGraph::Interface::PriorityQueue::Item    
                            
                        Item* GetNewItem() 
                            {return ((NULL != item_pool) ? item_pool->Get() : new Item);}
                        void InsertItem(Item& item);
                        void RemoveItem(Item& item)
                            {iface_list.Remove(item);}
                        
                        void PrependItem(Item& item)
                        {
                            item.SetQueue(this);
                            iface_list.Prepend(item);
                        }
                        void AppendItem(Item& item)
                        {   
                            item.SetQueue(this);
                            iface_list.Prepend(item);
                        }
                            
                        ProtoSortedTree     iface_list;  // sorted by "Cost" value
                        ItemPool*           item_pool;
                        
                        
                };  // end class ManetGraph::Interface::PriorityQueue


            private:
                Node&           node;
                ProtoAddress    address;
                Interface*      node_next;        // for Node's interface list

        };  // end class ManetGraph::Interface
        
        
        class Node
        {
            public:
                Node(const ProtoAddress& defaultAddr);
                ~Node();

                const ProtoAddress& GetDefaultAddress() const
                    {return default_interface.GetAddress();}

                Interface& GetDefaultInterface() {return default_interface;} 
                void AppendInterface(Interface& iface);
                void RemoveInterface(Interface& iface);

            private:
                ManetGraph::Interface   default_interface;

        };  // end class ManetGraph::Node
                
        // ManetGraph control/query methods
        void InsertNode(Node& node, Interface* iface = NULL);
        void RemoveNode(Node& node, Interface* iface = NULL);
        Interface* FindInterface(ProtoAddress& theAddress)
        {
            const char* key = theAddress.GetRawHostAddress();
            unsigned int keyBits = theAddress.GetLength() * 8;
            return static_cast<Interface*>(iface_tree.Find(key, keyBits));
        }
        
        Node* FindNode(ProtoAddress& theAddress)
        {
            Interface* iface = FindInterface(theAddress);
            return (NULL != iface) ? &iface->GetNode() : NULL;
        }
        
        class InterfaceIterator  : public ProtoGraph::VerticeIterator
        {
            public:
                InterfaceIterator(ManetGraph& theGraph);
                ~InterfaceIterator();

                void Reset() {ProtoGraph::VerticeIterator::Reset();}

                Interface* GetNextInterface()
                    {return static_cast<Interface*>(GetNextVertice());}
                
        };  // end class ManetGraph::InterfaceIterator
        
        class DijkstraTraversal
        {
            public:
                DijkstraTraversal(ManetGraph&  theGraph, 
                                  Node&        startNode,
                                  Interface*   startIface = NULL);
                ~DijkstraTraversal();
                
                bool Reset();
                
                Interface* GetNextInterface();
                
                void Update(Interface& startIface);
                
                
                // (use these post-Dijkstra (after "GetNextInterface" returns NULL)
                Interface* GetNextHop(const Interface& dstIface) const  // w/ "startIface" as srcIface
                    {return (queue_visited.GetNextHop(dstIface, start_iface));}
                Interface* GetPrevHop(const Interface& dstIface) const  // back towards "startIface"
                    {return (queue_visited.GetPrevHop(dstIface));}
                    
                const Cost* GetCost(const Interface& dstIface)
                    {return (queue_visited.GetCost(dstIface));}
                
                // BFS traversal of routing tree ("tree walk")
                bool TreeWalkReset();
                Interface* TreeWalkNext(unsigned int* level = NULL);
                
            private:
                ManetGraph&                 manet_graph;
                Interface&                  start_iface;
                Interface::PriorityQueue    queue_pending;
                Interface::PriorityQueue    queue_visited;
                
                // These two members support the "tree walk" BFS
                Interface*                  trans_iface;
                unsigned int                current_level;
                
                bool                        dijkstra_completed;
                bool                        in_update;
                
               
        };  // end class ManetGraph::DijkstraTraversal 
        
        class MinMaxTraversal
        {
            public:
                MinMaxTraversal(ManetGraph&  theGraph, 
                                  Node&        startNode,
                                  Interface*   startIface = NULL);
                ~MinMaxTraversal();
                
                bool Reset();
                
                Interface* GetNextInterface();
                
                void Update(Interface& startIface);
                
                
                // (use these post-Dijkstra (after "GetNextInterface" returns NULL)
                Interface* GetNextHop(const Interface& dstIface) const  // w/ "startIface" as srcIface
                    {return (queue_visited.GetNextHop(dstIface, start_iface));}
                Interface* GetPrevHop(const Interface& dstIface) const  // back towards "startIface"
                    {return (queue_visited.GetPrevHop(dstIface));}
                    
                const Cost* GetCost(const Interface& dstIface)
                    {return (queue_visited.GetCost(dstIface));}
                
                // BFS traversal of routing tree ("tree walk")
                bool TreeWalkReset();
                Interface* TreeWalkNext(unsigned int* level = NULL);
                
            private:
                ManetGraph&                 manet_graph;
                Interface&                  start_iface;
                Interface::PriorityQueue    queue_pending;
                Interface::PriorityQueue    queue_visited;
                
                // These two members support the "tree walk" BFS
                Interface*                  trans_iface;
                unsigned int                current_level;
                
                bool                        traversal_completed;
                bool                        in_update;
            
        };  // end class ManetGraph::MinMaxTraversal
    
    private:
        // Members
        ProtoTree   iface_tree;  // used to look up ifaces/nodes by address        
};  // end class ManetGraph

#endif // _MANET_GRAPH
